"""
PyDetex
https://github.com/ppizarror/PyDetex

A python application that removes LaTeX code and transforms it to plain text for grammar checking or other things...
"""

import pydetex.version

__author__ = 'Pablo Pizarro R.'
__copyright__ = 'Copyright 2021 Pablo Pizarro R. @ppizarror'
__description__ = 'A python application that removes LaTeX code and transforms it to plain text for grammar checking or other things...'
__email__ = 'pablo@ppizarror.com'
__keywords__ = 'latex detex parser gui'
__license__ = 'MIT'
__module_name__ = 'pydetex'
__url__ = 'https://github.com/ppizarror/PyDetex'
__url_bug_tracker__ = 'https://github.com/ppizarror/PyDetex'
__url_documentation__ = 'https://github.com/ppizarror/PyDetex'
__url_source_code__ = 'https://github.com/ppizarror/PyDetex'
__version__ = pydetex.version.ver
